from django import forms
from django.core import validators
from django.forms.formsets import DELETION_FIELD_NAME

from .validators import validate_methods, validate_path


class LinkBackForm(forms.Form):
    link_back = forms.CharField(initial="../", widget=forms.HiddenInput, required=False)


class TestRulesForm(LinkBackForm):
    field_template_name = "admin/django_fast_iprestrict/test_rules_field.html"
    test_ip = forms.CharField(
        validators=[validators.validate_ipv46_address],
        required=False,
        help_text="leave empty to use current ip address",
    )
    test_path = forms.CharField(
        validators=[validate_path],
        required=False,
        help_text="leave empty to test ip addresses only",
    )
    test_method = forms.CharField(
        validators=[validate_methods],
        required=False,
        help_text="leave empty to don't use method testing",
    )
    test_ratelimit_group = forms.CharField(
        validators=[validators.MinLengthValidator(1)],
        required=False,
        help_text="leave empty to not simulate a query from ratelimit apply_fn",
    )

    @property
    def has_data(self) -> bool:
        return bool(set(self.changed_data).difference({"link_back"}))


class ManagedForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field in self.instance.managed_fields:
            field = self.fields.get(field)
            if field:
                field.disabled = True
                field.widget.attrs["title"] = "managed"

    def clean(self):
        ret = super().clean()
        if self.instance.managed_fields:
            # MUST pop deletion field for preventing deletion
            if ret.pop(DELETION_FIELD_NAME, False):
                raise forms.ValidationError("Cannot delete managed form")

        return ret



from posixpath import dirname

from django.conf import settings
from django.contrib import admin
from django.core.cache import caches
from django.core.exceptions import PermissionDenied
from django.db import models
from django.db.models.functions import Cast
from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseRedirect,
)
from django.template.response import TemplateResponse
from django.urls import path
from django.utils.html import format_html

from .forms import LinkBackForm, ManagedForm, TestRulesForm
from .models import (
    Rule,
    RuleNetwork,
    RulePath,
    RuleRatelimit,
    RuleRatelimitGroup,
    RuleSource,
)
from .utils import RULE_ACTION, LockoutException, get_ip, parse_ipaddress

try:
    import django_fast_ratelimit as ratelimit
except ImportError:
    ratelimit = None


class IsRatelimitMatcherFilter(admin.SimpleListFilter):
    title = "is ratelimit matcher"
    parameter_name = "is_ratelimit_matcher"

    def lookups(self, request, model_admin):
        return [("0", "Normal mode"), ("1", "Ratelimit matcher mode")]

    def queryset(self, request, queryset):
        value = self.value()
        if value == "0":
            return queryset.filter(
                ~models.Exists(
                    RuleRatelimitGroup.objects.filter(rule_id=models.OuterRef("id"))
                )
            )
        elif value == "1":
            return queryset.filter(
                models.Exists(
                    RuleRatelimitGroup.objects.filter(rule_id=models.OuterRef("id"))
                )
            )
        return queryset


class IsManagedFilter(admin.SimpleListFilter):
    title = "is managed"
    parameter_name = "is_managed"

    def lookups(self, request, model_admin):
        return [("0", "Unmanaged"), ("1", "Managed")]

    def queryset(self, request, queryset):
        value = self.value()
        if value == "0":
            return queryset.alias(
                managed_fields_as_str=Cast("managed_fields", models.TextField())
            ).filter(managed_fields_as_str="[]")
        elif value == "1":
            return queryset.alias(
                managed_fields_as_str=Cast("managed_fields", models.TextField())
            ).exclude(managed_fields_as_str="[]")
        return queryset


class ManageableMixin:
    form = ManagedForm

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = list(super().get_readonly_fields(request, obj))
        if obj and isinstance(obj, self.model):
            readonly_fields.extend(obj.managed_fields)
        return readonly_fields

    def has_delete_permission(self, request, obj=None):
        if not super().has_delete_permission(request, obj):
            return False
        if not obj:
            return True
        if isinstance(obj, self.model):
            # can only delete without managed fields
            return not bool(obj.managed_fields)
        opts = self.opts
        parent_field = opts.get_field("rule")
        return parent_field.remote_field.name not in obj.managed_fields

    @admin.display(
        description="managed",
        boolean=True,
    )
    def is_managed(self, obj):
        return bool(obj.managed_fields)

    @admin.display(description="managed fields")
    def managed_fields_display(self, obj):
        return ", ".join(obj.managed_fields)


class TestRulesMixin:
    def changelist_view(self, request, extra_context=None):
        default = {
            "test_rules_form": TestRulesForm(
                data={
                    **request.session.get("iprestrict_test_rules_data", {}),
                    "test_rules_form-link_back": request.path,
                },
                prefix="test_rules_form",
            )
        }
        if extra_context:
            default.update(extra_context)
        return super().changelist_view(
            request,
            extra_context=default,
        )


class RuleSubMixin(TestRulesMixin, ManageableMixin):
    if hasattr(admin, "ShowFacets"):
        show_facets = admin.ShowFacets.ALWAYS
    list_filter = (
        "is_active",
        IsManagedFilter,
    )

    @admin.display(ordering="rule__name", description="Rule")
    def rule_display(self, obj):
        return obj.rule.name

    def get_changelist_form(self, request, **kwargs):
        return super().get_changelist_form(request, form=ManagedForm, **kwargs)

    def get_urls(self):
        return [
            path("<str:object_id>/change/", self.redirect_change),
            *super().get_urls(),
        ]

    def redirect_change(self, request, object_id):
        subruleobj = self.model.objects.get(id=object_id)
        return HttpResponseRedirect(f"../../../rule/{subruleobj.rule_id}/change/")

    def has_add_permission(self, request):
        return False


class SubRuleInlineMixin(ManageableMixin):
    extra = 1
    fk_name = "rule"

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = list(super().get_readonly_fields(request, obj))
        readonly_fields.append("is_managed")
        return readonly_fields

    def get_extra(self, request, obj=None, **kwargs):
        return self.extra if not obj else 0

    def has_add_permission(self, request, obj=None):
        if not super().has_add_permission(request, obj):
            return False
        if not obj:
            return True
        if isinstance(obj, self.model):
            return True
        # check parent model
        opts = self.opts
        parent_field = opts.get_field("rule")
        return parent_field.remote_field.name not in obj.managed_fields


class RulePathInlineAdmin(SubRuleInlineMixin, admin.StackedInline):
    model = RulePath


class RuleRatelimitInlineAdmin(SubRuleInlineMixin, admin.StackedInline):
    model = RuleRatelimit


class RuleNetworkInlineAdmin(SubRuleInlineMixin, admin.TabularInline):
    model = RuleNetwork


class RuleSourceInlineAdmin(SubRuleInlineMixin, admin.TabularInline):
    model = RuleSource


class RuleRatelimitGroupInlineAdmin(SubRuleInlineMixin, admin.TabularInline):
    model = RuleRatelimitGroup


# reuse the existing form to prevent form nesting
_position_template = """
<div style="display:flex; gap: 16px" >
    <button
        style="border-radius: 30%;"
        name="rule_move_direction"
        title="start"
        value="start"
        formaction="{object_id}/rule_move/"
    >
        &#x21C8;
    </button>
    <button
        style="border-radius: 30%;"
        name="rule_move_direction"
        title="up" value="up"
        formaction="{object_id}/rule_move/"
    >
        &#x2191;
    </button>
    <button
        style="border-radius: 30%;"
        name="rule_move_direction"
        title="down" value="down"
        formaction="{object_id}/rule_move/"
    >
        &#x2193;
    </button>
    <button
        style="border-radius: 30%;"
        name="rule_move_direction"
        title="end"
        value="end"
        formaction="{object_id}/rule_move/"
    >
        &#x21CA;
    </button>
</div>
"""


@admin.register(Rule)
class RuleAdmin(TestRulesMixin, ManageableMixin, admin.ModelAdmin):
    if hasattr(admin, "ShowFacets"):
        show_facets = admin.ShowFacets.ALWAYS
    list_display = (
        "position_short",
        "position_buttons",
        "action",
        "name",
        "methods",
        "invert_methods",
        "is_ratelimit_matcher",
        "is_managed",
    )
    list_display_links = ("position_short",)
    list_editable = ("name", "action", "methods", "invert_methods")
    ordering = ("position",)
    fields = ["name", "managed_fields_display", "action", "methods", "invert_methods"]
    list_filter = (
        "action",
        IsManagedFilter,
        IsRatelimitMatcherFilter,
    )
    search_fields = ("name",)
    inlines = [
        RuleNetworkInlineAdmin,
        RuleSourceInlineAdmin,
        RulePathInlineAdmin,
        RuleRatelimitInlineAdmin,
        RuleRatelimitGroupInlineAdmin,
    ]

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = list(super().get_readonly_fields(request, obj))
        readonly_fields.append("managed_fields_display")
        return readonly_fields

    def get_changelist_form(self, request, **kwargs):
        return super().get_changelist_form(request, form=ManagedForm, **kwargs)

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .annotate(
                has_ratelimit_groups=models.Exists(
                    RuleRatelimitGroup.objects.filter(
                        is_active=True, rule_id=models.OuterRef("id")
                    )
                ),
            )
        )

    def get_urls(self):
        return [
            path("test_rules/", self.test_rules),
            path("clear_caches/", self.clear_iprestrict_caches),
            path(
                "simulate_rules<path:test_path>",
                self.simulate_rules,
            ),
            path("<str:object_id>/rule_move/", self.rule_move),
            *super().get_urls(),
        ]

    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        try:
            return super().changeform_view(
                request,
                object_id=object_id,
                form_url=form_url,
                extra_context=extra_context,
            )
        except LockoutException:
            context = {**self.admin_site.each_context(request), "link_back": "./"}
            return TemplateResponse(
                request,
                "admin/django_fast_iprestrict/lockout_prevented.html",
                context=context,
                status=400,
            )

    def changelist_view(self, request, extra_context=None):
        try:
            return super().changelist_view(
                request,
                extra_context=extra_context,
            )
        except LockoutException:
            context = {**self.admin_site.each_context(request), "link_back": "./"}
            return TemplateResponse(
                request,
                "admin/django_fast_iprestrict/lockout_prevented.html",
                context=context,
                status=400,
            )

    def save_model(self, request, obj, form, change):
        obj._trigger_cleanup = False
        super().save_model(request, obj, form, change)

    def delete_model(self, request, obj):
        obj._trigger_cleanup = False
        super().delete_model(request, obj)
        Rule.objects.position_cleanup(ip=get_ip(request), path=request.path)

    def delete_queryset(self, request, queryset):
        super().delete_queryset(request, queryset)
        parent_path = dirname(request.path.rstrip("/"))

        Rule.objects.position_cleanup(ip=get_ip(request), path=parent_path)

    def save_related(self, request, form, formsets, change):
        super().save_related(request, form, formsets, change)
        Rule.objects.position_cleanup(ip=get_ip(request), path=request.path)

    def rule_move(self, request: HttpRequest, object_id):
        if request.method != "POST":
            # fail soft for browsers visiting this url
            return HttpResponseRedirect("../../")
        direction = request.POST["rule_move_direction"]
        parent_path = dirname(request.path.rstrip("/"))
        try:
            if direction == "up":
                self.model.objects.position_up(
                    object_id, ip=get_ip(request), path=parent_path
                )
            elif direction == "down":
                self.model.objects.position_down(
                    object_id, ip=get_ip(request), path=parent_path
                )
            elif direction == "start":
                self.model.objects.position_start(
                    object_id, ip=get_ip(request), path=parent_path
                )
            elif direction == "end":
                self.model.objects.position_end(
                    object_id, ip=get_ip(request), path=parent_path
                )
            else:
                return HttpResponseBadRequest()
        except LockoutException:
            context = {**self.admin_site.each_context(request), "link_back": "../../"}
            return TemplateResponse(
                request,
                "admin/django_fast_iprestrict/lockout_prevented.html",
                context=context,
                status=400,
            )
        return HttpResponseRedirect("../../")

    def simulate_rules(self, request, test_path):
        rule_id, action, is_catch_all, ratelimits = RulePath.objects.match_ip_and_path(
            get_ip(request), test_path
        )

        if ratelimit:
            for rdict in ratelimits:
                r = ratelimit.get_ratelimit(
                    request=request,
                    # don't use action
                    action=ratelimit.Action.PEEK,
                    group=rdict["group"],
                    key=rdict["key"],
                    rate=rdict["rate"],
                )
                r.decorate_object(
                    request, name=rdict["decorate_name"], block=rdict["block"]
                )
        if action == RULE_ACTION.deny:
            raise PermissionDenied()

        return HttpResponse(f"accessed: {test_path}")

    def test_rules(self, request):
        from django.contrib.messages import ERROR, INFO, SUCCESS, WARNING

        if request.method != "POST":
            # fail soft for browsers visiting this url
            return HttpResponseRedirect("../")
        form = TestRulesForm(
            request.POST,
            prefix="test_rules_form",
        )
        if form.is_valid():
            test_method = request.POST.get("test_method", None) or None
            test_ratelimit_group = (
                form.cleaned_data.get("test_ratelimit_group", None) or None
            )

            test_ip = form.cleaned_data.get("test_ip", None)
            if not test_ip:
                test_ip = get_ip(request)

            test_path = form.cleaned_data.get("test_path", None) or ""
            if test_path:
                rule_id = RulePath.objects.match_ip_and_path(
                    ip=test_ip,
                    path=test_path,
                    method=test_method,
                    ratelimit_group=test_ratelimit_group,
                )[0]
                self.message_user(
                    request,
                    f"Parameters: ip: {parse_ipaddress(test_ip)}, path: {test_path}, method: {test_method or '-'}, ratelimit group: {test_ratelimit_group or '-'}",
                    level=INFO,
                )
            else:
                rule_id = Rule.objects.match_ip(
                    ip=test_ip,
                    method=test_method,
                    ratelimit_group=test_ratelimit_group,
                )[0]
                self.message_user(
                    request,
                    f"Parameters: ip: {parse_ipaddress(test_ip)}, method: {test_method or '-'}, ratelimit group: {test_ratelimit_group or '-'}",
                    level=INFO,
                )
            if rule_id:
                rule = Rule.objects.get(id=rule_id)
                self.message_user(
                    request,
                    f"Matched rule: {rule.name}, action: {RULE_ACTION(rule.action).label}",
                    level=SUCCESS
                    if rule.action == RULE_ACTION.allow.value
                    else WARNING,
                )
            else:
                self.message_user(request, "No rule matched", level=ERROR)

        session_data = {}

        for key in form.fields:
            prefixed_key = f"test_rules_form-{key}"
            if key != "link_back" and prefixed_key in form.data:
                session_data[prefixed_key] = form.data[prefixed_key]
        request.session["iprestrict_test_rules_data"] = session_data
        return HttpResponseRedirect(form.cleaned_data.get("link_back", "../"))

    def clear_iprestrict_caches(self, request):
        if request.method != "POST":
            # fail soft for browsers visiting this url
            raise HttpResponseRedirect("../")
        form = LinkBackForm(
            request.POST,
            prefix="test_rules_form",
        )
        if form.is_valid():
            # set => deduplicating
            for cache_name in {
                getattr(settings, "IPRESTRICT_CACHE", "default"),
                getattr(settings, "RATELIMIT_CACHE", "default"),
            }:
                caches[cache_name].clear()
        return HttpResponseRedirect(f"{form.cleaned_data.get('link_back', '../')}")

    @admin.display(description="")
    def position_short(self, obj):
        return obj.position

    @admin.display(ordering="position", description="Position")
    def position_buttons(self, obj):
        return format_html(
            _position_template,
            object_id=obj.id,
        )

    @admin.display(
        ordering="has_ratelimit_groups",
        description="is ratelimit matcher",
        boolean=True,
    )
    def is_ratelimit_matcher(self, obj):
        return obj.has_ratelimit_groups


@admin.register(RulePath)
class RulePathAdmin(RuleSubMixin, admin.ModelAdmin):
    list_display = ("rule_display", "path", "is_regex", "is_active", "is_managed")
    ordering = ("rule__name", "path")
    search_fields = ("path", "rule__name")
    # cannot enable list editable yet, needs lockout check


@admin.register(RuleNetwork)
class RuleNetworkAdmin(RuleSubMixin, admin.ModelAdmin):
    list_display = ("rule_display", "network", "is_active", "is_managed")
    ordering = ("rule__name", "network")
    search_fields = ("network", "rule__name")
    # cannot enable list editable yet, needs lockout check


@admin.register(RuleRatelimit)
class RuleRatelimitAdmin(RuleSubMixin, admin.ModelAdmin):
    list_display = (
        "rule_display",
        "group",
        "key",
        "rate",
        "action",
        "block",
        "wait",
        "is_active",
        "is_managed",
    )
    ordering = ("rule__name", "group")
    list_filter = (*RuleSubMixin.list_filter, "action")
    list_editable = (
        "group",
        "key",
        "rate",
        "action",
        "block",
        "wait",
        "is_active",
    )
    search_fields = ("group", "key", "rule__name")


@admin.register(RuleRatelimitGroup)
class RuleRatelimitGroupAdmin(RuleSubMixin, admin.ModelAdmin):
    list_display = ("rule_display", "name", "is_active", "is_managed")
    ordering = ("rule__name", "name")
    list_editable = ("name", "is_active")
    search_fields = ("name", "rule__name")


@admin.register(RuleSource)
class RuleSourceAdmin(RuleSubMixin, admin.ModelAdmin):
    list_display = (
        "rule_display",
        "generator_fn",
        "interval",
        "is_active",
        "is_managed",
    )
    ordering = ("rule__name", "generator_fn")
    list_editable = ("generator_fn", "interval", "is_active")
    search_fields = ("generator_fn", "rule__name")



# Generated by Django 5.0.1 on 2024-01-26 20:42

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_fast_iprestrict", "0006_rule_invert_methods_rule_methods_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="RuleRatelimitGroup",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        verbose_name="ratelimit group name",
                        help_text="matcher for django-fast-ratelimit group",
                        max_length=80,
                        validators=[django.core.validators.MinLengthValidator(1)],
                    ),
                ),
                ("is_active", models.BooleanField(blank=True, default=True)),
                (
                    "rule",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ratelimit_groups",
                        to="django_fast_iprestrict.rule",
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="ruleratelimitgroup",
            constraint=models.UniqueConstraint(
                fields=("rule_id", "name"), name="rule_ratelimit_groups_unique"
            ),
        ),
    ]



from django.core.cache import caches
from django.core.management.base import BaseCommand

from ...models import Rule
from ...utils import RULE_ACTION


class Command(BaseCommand):
    help = "Emergency disable all rules and clear caches for reallowing login"

    def handle(self, **kwargs):
        Rule.objects.filter(action=RULE_ACTION.deny.value).update(
            action=RULE_ACTION.disabled.value
        )
        for cache in caches.all():
            cache.clear()
            cache.close()

        # clearing local_caches doesn't help, use different proccess



__all__ = ["apply_iprestrict"]

import asyncio
from functools import partial, singledispatch
from typing import Optional, Union

from django.http import HttpRequest

from .utils import RATELIMIT_ACTION, RULE_ACTION, get_ip

try:
    import django_fast_ratelimit as ratelimit
except ImportError:
    ratelimit = None

if ratelimit:

    def _apply_iprestrict(
        request,
        group,
        action=None,
        rate=None,
        ignore_pathes=False,
        require_rule=False,
        no_count=False,
        no_execute=False,
        default_action: Optional[Union[str, RULE_ACTION]] = None,
    ):
        # don't check methods here as the check is done in ratelimit
        from .models import Rule, RulePath, RuleRatelimitGroup

        if action == RATELIMIT_ACTION.RESET or action == RATELIMIT_ACTION.RESET_EPOCH:
            no_count = True
            no_execute = True

        ip = get_ip(request)
        with_path = False
        if not ignore_pathes:
            name_has_pathes = RuleRatelimitGroup.objects.name_matchers()[1]
            with_path = group in name_has_pathes
        if with_path:
            rule_id, iprestrict_action, _, ratelimits = (
                RulePath.objects.match_ip_and_path(
                    ip=ip,
                    path=request.path,
                    ratelimit_group=group,
                    default_action=default_action,
                )
            )
        else:
            rule_id, iprestrict_action, _, ratelimits = Rule.objects.match_ip(
                ip=ip,
                ratelimit_group=group,
                default_action=default_action,
            )
        if rule_id is None and require_rule:
            raise ratelimit.Disabled(
                "no rule found for %s" % group,
                ratelimit=ratelimit.Ratelimit(group=group, request_limit=1, end=0),
            )

        for rdict in ratelimits:
            if rdict["rate"] == "inherit":
                r_rate = rate
                if not r_rate:
                    continue
            else:
                r_rate = rdict["rate"]
            r = ratelimit.get_ratelimit(
                request=request,
                action=ratelimit.Action.PEEK if no_count else rdict["action"],
                group=rdict["group"],
                key=rdict["key"],
                rate=r_rate,
            )
            r.decorate_object(
                request,
                name=rdict["decorate_name"],
                block=rdict["block"] and not no_execute,
            )
            if action == RATELIMIT_ACTION.RESET:
                r.reset()
            elif action == RATELIMIT_ACTION.RESET_EPOCH:
                r.reset(request)
        if no_execute and not default_action:
            return 0
        if iprestrict_action == RULE_ACTION.deny:
            return 1
        return 0

    async def _aapply_iprestrict(
        request,
        group,
        action=None,
        rate=None,
        ignore_pathes=False,
        require_rule=False,
        no_count=False,
        no_execute=False,
        default_action: Optional[Union[str, RULE_ACTION]] = None,
    ):
        # don't check methods here as the check is done in ratelimit
        from .models import Rule, RulePath, RuleRatelimitGroup

        if action == RATELIMIT_ACTION.RESET or action == RATELIMIT_ACTION.RESET_EPOCH:
            no_count = True
            no_execute = True

        ip = get_ip(request)

        with_path = False
        if not ignore_pathes:
            name_has_pathes = (await RuleRatelimitGroup.objects.aname_matchers())[1]
            with_path = group in name_has_pathes
        if with_path:
            (
                rule_id,
                iprestrict_action,
                _,
                ratelimits,
            ) = await RulePath.objects.amatch_ip_and_path(
                ip=ip,
                path=request.path,
                ratelimit_group=group,
                default_action=default_action,
            )
        else:
            rule_id, iprestrict_action, _, ratelimits = await Rule.objects.amatch_ip(
                ip=ip,
                ratelimit_group=group,
                default_action=default_action,
            )
        if rule_id is None and require_rule:
            raise ratelimit.Disabled(
                "no rule found for %s" % group,
                ratelimit=ratelimit.Ratelimit(group=group, request_limit=1, end=0),
            )
        for rdict in ratelimits:
            if rdict["rate"] == "inherit":
                r_rate = rate
                if not r_rate:
                    continue
            else:
                r_rate = rdict["rate"]
            r = await ratelimit.aget_ratelimit(
                request=request,
                action=ratelimit.Action.PEEK if no_count else rdict["action"],
                group=rdict["group"],
                key=rdict["key"],
                rate=r_rate,
            )
            await r.adecorate_object(
                request,
                name=rdict["decorate_name"],
                wait=rdict["wait"] and not no_execute,
                block=rdict["block"] and not no_execute,
            )
            if action == RATELIMIT_ACTION.RESET:
                await r.areset()
            elif action == RATELIMIT_ACTION.RESET_EPOCH:
                await r.areset(request)
        if no_execute and not default_action:
            return 0
        if iprestrict_action == RULE_ACTION.deny:
            return 1
        return 0

    @singledispatch
    def apply_iprestrict(request, group, action=None, rate=None, **kwargs):
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None
        if loop:
            return _aapply_iprestrict(request, group, action, rate, **kwargs)
        else:
            return _apply_iprestrict(request, group, action, rate, **kwargs)

    @apply_iprestrict.register(str)
    @apply_iprestrict.register(list)
    @apply_iprestrict.register(tuple)
    def _(arg="", *args):
        args = list(args)
        if isinstance(arg, str):
            args.extend(arg.split(","))
        default_action = None
        for arg in args:
            if isinstance(arg, str):
                if arg.startswith("default_action:"):
                    default_action = arg.split(":", 1)[-1]
            elif isinstance(arg, (tuple, list)) and len(arg) >= 2:
                if arg[0] == "default_action":
                    # call to verify if already the right value/format
                    default_action = arg[1]
        return partial(
            apply_iprestrict.dispatch(HttpRequest),
            ignore_pathes="ignore_pathes" in args,
            require_rule="require_rule" in args,
            no_count="no_count" in args or "execute_only" in args,
            no_execute="no_execute" in args or "count_only" in args,
            default_action=default_action,
        )

else:

    def apply_iprestrict(*args, **kwargs):
        raise ImportError("ratelimit not installed")



# Generated by Django 5.0.1 on 2024-01-30 19:43

from django.db import migrations, models

import django_fast_iprestrict.validators


class Migration(migrations.Migration):
    dependencies = [
        ("django_fast_iprestrict", "0009_alter_ruleratelimit_rate"),
    ]

    operations = [
        migrations.AlterField(
            model_name="ruleratelimit",
            name="rate",
            field=models.CharField(
                help_text=(
                    'Set to "inherit" to use the provided rate (when set to "inherit" '
                    "ratelimit is ignored in middleware and without a provided rate)"
                ),
                max_length=10,
                validators=[django_fast_iprestrict.validators.validate_rate],
            ),
        ),
    ]



# Generated by Django 5.0.1 on 2024-01-30 19:05

import django_fast_iprestrict.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_fast_iprestrict', '0008_add_managed_fields'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ruleratelimit',
            name='rate',
            field=models.CharField(help_text='Leave empty to use the default rate (only valid for ratelimit matching rules)', max_length=10, validators=[django_fast_iprestrict.validators.validate_rate]),
        ),
    ]



from django.conf import settings
from django.core.cache import caches
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "Clear caches"

    def add_arguments(self, parser):
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "-n",
            "--names",
            nargs="+",
            default=(
                getattr(settings, "IPRESTRICT_CACHE", "default"),
                getattr(settings, "RATELIMIT_CACHE", "default"),
            ),
        )
        group.add_argument("--all", action="store_true", dest="clear_all")

    def handle(self, clear_all, names, **kwargs):
        for cache_name in set(caches if clear_all else names):
            print("clear", cache_name)
            cache = caches[cache_name]
            cache.clear()
        caches.close_all()



import logging
import re
import time
from contextlib import ExitStack
from functools import lru_cache
from typing import Optional

from asgiref.sync import sync_to_async
from django.conf import settings
from django.core.cache import caches
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Max, Window
from django.db.models.functions import Concat, RowNumber
from django.db.transaction import atomic
from django.utils.module_loading import import_string

from .utils import (
    RATELIMIT_ACTION,
    RULE_ACTION,
    LockoutException,
    get_default_action,
    get_default_interval,
    invertedset,
    parse_ipaddress,
    parse_ipnetwork,
)
from .validators import (
    min_length_1,
    validate_generator_fn,
    validate_methods,
    validate_network,
    validate_path,
    validate_rate,
    validate_ratelimit_key,
    validate_regex,
)

logger = logging.getLogger(__name__)
# Create your models here.

_empty = ()
_update_interval_secs = 2 * 60


def get_default_position():
    return Rule.objects.aggregate(Max("position", default=0))["position__max"] + 1


class Manageable(models.Model):
    managed_fields = models.JSONField(blank=True, default=list, editable=False)

    class Meta:
        abstract = True

    def clean(self):
        super().clean()
        valid_names = {i.name for i in self._meta.get_fields()}
        valid_names.discard("position")
        self.managed_fields = list(valid_names.intersection(self.managed_fields))


class ActivatableAndManageable(Manageable):
    is_active = models.BooleanField(blank=True, default=True)

    class Meta:
        abstract = True


class RuleManager(models.Manager):
    _next_rules_updates = time.monotonic() + _update_interval_secs

    def _atomic_update(self, ip=None, path=None, use_atomic=True):
        stack = ExitStack()
        stack.queryset = self.all().select_for_update()
        stack.enter_context(atomic())
        if ip:
            stack.callback(
                self.lockout_check, ip=ip, path=path, clear_caches_on_error=True
            )

        stack.callback(self._position_cleanup, stack.queryset)
        return stack

    def clear_local_caches(self):
        RulePath.objects.path_matchers.cache_clear()
        RuleRatelimitGroup.objects.name_matchers.cache_clear()
        self._ip_matchers_local.cache_clear()
        self._next_rules_updates = time.monotonic() + _update_interval_secs

    def lockout_check(self, ip, path=None, clear_caches_on_error=True, remote=False):
        if not ip:
            return
        if path:
            if (
                RulePath.objects.match_ip_and_path(ip=ip, path=path, remote=remote)[1]
                == RULE_ACTION.deny
            ):
                # otherwise the caches could cause access errors until server restart
                if clear_caches_on_error:
                    self.clear_local_caches()
                raise LockoutException("would lock current user out")

        else:
            if self.match_ip(ip=ip, remote=remote)[1] == RULE_ACTION.deny:
                # otherwise the caches could cause access errors until server restart
                if clear_caches_on_error:
                    self.clear_local_caches()
                raise LockoutException("would lock current user out")

    def _position_cleanup(self, queryset):
        u_dict = {}
        # window functions don't like select_for_update
        for val in self.all().annotate(
            position_new=Window(expression=RowNumber(), order_by="position"),
        ):
            if val.position != val.position_new:
                u_dict[val.id] = val.position_new
        for rule_id, position in u_dict.items():
            queryset.filter(id=rule_id).update(position=position)
        self.clear_local_caches()

    def position_cleanup(self, ip=None, path=None):
        with self._atomic_update(ip=ip, path=path):
            pass

    @lru_cache(maxsize=2)
    def _ip_matchers_local(self, generic=False):
        # ordered dict
        patterns = {}
        name_matchers = RuleRatelimitGroup.objects.name_matchers()[0]
        queryset = self.annotate(
            has_networks=models.Exists(
                RuleNetwork.objects.filter(
                    is_active=True, rule_id=models.OuterRef("id")
                )
            ),
            has_sources=models.Exists(
                RuleSource.objects.filter(is_active=True, rule_id=models.OuterRef("id"))
            ),
            has_pathes=models.Exists(
                RulePath.objects.filter(is_active=True, rule_id=models.OuterRef("id"))
            ),
        )
        if not generic:
            # non catchalls are excluded, when without network/source
            queryset = queryset.exclude(
                models.Q(has_sources=False, has_networks=False, has_pathes=True)
            ).exclude(action=RULE_ACTION.disabled.value)
        for obj in queryset.distinct():
            patterns[obj.id] = (
                ["*"] if obj._is_catchall() else obj.get_processed_networks(),  # 0
                obj.get_methods(),  # 1
                name_matchers.get(
                    obj.id, _empty
                ),  # 2, note MUST be _empty otherwise the detection if a RuleRatelimitGroup is just disabled fails
                RULE_ACTION(obj.action),  # 3
                obj._is_catchall(),  # 4
                obj.get_ratelimit_dicts(),  # 5
            )
        return patterns

    def ip_matchers_local(self, generic=False):
        if self._next_rules_updates < time.monotonic():
            self.clear_local_caches()
        return self._ip_matchers_local(generic=generic)

    aip_matchers_local = sync_to_async(ip_matchers_local)

    def match_ip(
        self,
        ip: str,
        method: Optional[str] = None,
        ratelimit_group: Optional[str] = None,
        rule_id: Optional[int] = None,
        generic=False,
        remote=True,
        # overwrite default action
        default_action=None,
    ):
        ip_address_user = parse_ipaddress(ip)
        if rule_id:
            item = self.ip_matchers_local(generic=generic).get(rule_id, None)
            if not item:
                return None, get_default_action(default_action), False, _empty
            if method and method not in item[1]:
                return None, get_default_action(default_action), False, _empty

            if ratelimit_group:
                if ratelimit_group not in item[2]:
                    return None, get_default_action(default_action), False, _empty
            elif item[2] is not _empty:
                return None, get_default_action(default_action), False, _empty

            for network in item[0]:
                try:
                    if network == "*" or ip_address_user in network:
                        return rule_id, *item[3:]
                except TypeError:
                    pass
            if remote:
                for remote_networks in RuleSource.objects.ip_matchers_remote(
                    [rule_id]
                ).values():
                    # maximal one
                    for network in remote_networks:
                        try:
                            if ip_address_user in network:
                                return rule_id, *item[3:]
                        except TypeError:
                            pass

        else:
            ip_matchers = self.ip_matchers_local(generic=generic)
            ip_matchers_remote = (
                RuleSource.objects.ip_matchers_remote(ip_matchers.keys())
                if remote
                else {}
            )
            for rule_id, item in ip_matchers.items():
                if method and method not in item[1]:
                    continue
                if ratelimit_group:
                    if ratelimit_group not in item[2]:
                        continue
                elif item[2] is not _empty:
                    continue

                for network in item[0]:
                    try:
                        if network == "*" or ip_address_user in network:
                            return rule_id, *item[3:]
                    except TypeError:
                        pass

                for network in ip_matchers_remote.get(str(rule_id), _empty):
                    try:
                        if ip_address_user in network:
                            return rule_id, *item[3:]
                    except TypeError:
                        pass
        return None, get_default_action(default_action), False, _empty

    amatch_ip = sync_to_async(match_ip)

    def match_all_ip(
        self,
        ip: str,
        method: Optional[str] = None,
        ratelimit_group=None,
        remote=True,
        generic=False,
    ):
        # note: cannot apply ratelimit here
        ip_address_user = parse_ipaddress(ip)
        result = []
        ip_matchers = self.ip_matchers_local(generic=generic)
        ip_matchers_remote = None
        for rule_id, item in ip_matchers.items():
            if method and method not in item[1]:
                continue
            if ratelimit_group:
                if ratelimit_group not in item[2]:
                    continue
            elif item[2] is not _empty:
                continue
            found_local_network = False
            for network in item[0]:
                try:
                    if network == "*" or ip_address_user in network:
                        result.append((rule_id, *item[3:]))
                        found_local_network = True
                        break
                except TypeError:
                    pass
            if found_local_network:
                continue
            # lazy fetch
            if ip_matchers_remote is None:
                ip_matchers_remote = (
                    RuleSource.objects.ip_matchers_remote(ip_matchers.keys())
                    if remote
                    else {}
                )
            for network in ip_matchers_remote.get(str(rule_id), _empty):
                try:
                    if ip_address_user in network:
                        result.append((rule_id, *item[2:]))
                        break
                except TypeError:
                    pass

        return result

    amatch_all_ip = sync_to_async(match_all_ip)

    def position_start(self, rule_id, ip=None, path=None):
        with self._atomic_update(ip=ip, path=path) as stack:
            stack.queryset.filter(id=rule_id).update(position=0)

    def position_end(self, rule_id, ip=None, path=None):
        with self._atomic_update(ip=ip, path=path) as stack:
            max_position = stack.queryset.aggregate(
                max_position=Max("position", default=0)
            )["max_position"]
            # ugly hack
            self.get_queryset().annotate().filter(id=rule_id).update(
                position=max_position + 1
            )

    def position_up(self, rule_id, ip=None, path=None):
        with self._atomic_update(ip=ip, path=path) as stack:
            rule_position = stack.queryset.filter(id=rule_id).values("position")
            stack.queryset.annotate(
                position_mod=models.Case(
                    models.When(position__lte=1, id=rule_id, then=models.Value(0)),
                    models.When(id=rule_id, then=models.Value(-1)),
                    models.When(
                        position=models.Subquery(rule_position) - models.Value(1),
                        then=models.Value(1),
                    ),
                    default=models.Value(0),
                ),
            ).update(position=models.F("position") + models.F("position_mod"))

    def position_down(self, rule_id, ip=None, path=None):
        with self._atomic_update(ip=ip, path=path) as stack:
            rule_position = stack.queryset.filter(id=rule_id).values("position")
            stack.queryset.annotate(
                position_mod=models.Case(
                    models.When(id=rule_id, then=models.Value(1)),
                    models.When(
                        position=models.Subquery(rule_position) + models.Value(1),
                        then=models.Value(-1),
                    ),
                    default=models.Value(0),
                ),
            ).update(position=models.F("position") + models.F("position_mod"))


class Rule(Manageable):
    _trigger_cleanup = True
    position = models.PositiveIntegerField(blank=True, default=get_default_position)
    name = models.CharField(max_length=50, unique=True)
    methods = models.CharField(
        max_length=100,
        validators=[validate_methods],
        default="",
        blank=True,
        help_text="comma seperated http methods",
    )
    invert_methods = models.BooleanField(
        default=True,
        blank=True,
        help_text="exclude instead of include http methods. Set and leave methods empty to match all",
    )
    action = models.CharField(
        max_length=1, choices=RULE_ACTION.choices, default=RULE_ACTION.disabled
    )
    objects = RuleManager()

    class Meta:
        ordering = ("position",)

        indexes = [
            models.Index(fields=["position"]),
        ]

    def __str__(self):
        return self.name

    def is_catch_all(self, also_disabled=False):
        rule = type(self).objects.ip_matchers_local(generic=also_disabled).get(self.id)
        return bool(rule and rule[3])

    def _is_catchall(self):
        # use annotations
        return not self.has_networks and not self.has_sources

    def get_processed_networks(self):
        return [
            parse_ipnetwork(network)
            for network in self.networks.filter(is_active=True).values_list(
                "network", flat=True
            )
        ]

    def get_ratelimit_dicts(self):
        return list(
            self.ratelimits.filter(is_active=True).values(
                "key", "group", "rate", "decorate_name", "block", "wait", "action"
            )
        )

    def get_methods(self):
        if self.invert_methods:
            if not self.methods:
                return invertedset()
            return invertedset(self.methods.split(","))
        if not self.methods:
            return set()
        return set(self.methods.split(","))

    def match_ip(self, ip, method=None, remote=True, generic=False):
        return type(self).objects.match_ip(
            ip, method=method, rule_id=self.id, remote=remote, generic=generic
        )

    amatch_ip = sync_to_async(match_ip)

    def clean(self):
        super().clean()
        self.methods = ",".join(
            sorted(map(lambda x: x.upper().strip(), self.methods.split(",")))
        )


class RuleRatelimitManager(models.Manager):
    pass


class RuleRatelimit(ActivatableAndManageable):
    rule = models.ForeignKey(Rule, related_name="ratelimits", on_delete=models.CASCADE)
    key = models.CharField(
        max_length=200,
        validators=[
            validate_ratelimit_key,
        ],
    )
    group = models.CharField(max_length=50)
    decorate_name = models.CharField(max_length=50, default="ratelimit", blank=True)
    action = models.SmallIntegerField(
        choices=RATELIMIT_ACTION.choices,
        default=RATELIMIT_ACTION.INCREASE,
    )
    rate = models.CharField(
        max_length=10,
        validators=[validate_rate],
        help_text=(
            'Set to "inherit" to use the provided rate (when set to "inherit" '
            "ratelimit is ignored in middleware and without a provided rate)"
        ),
    )
    block = models.BooleanField(blank=True, default=False)
    wait = models.BooleanField(blank=True, default=False)

    objects = RuleRatelimitManager()


class RuleRatelimitGroupManager(models.Manager):
    @lru_cache(maxsize=1)
    def name_matchers(self):
        rule_id_to_names = {}
        name_has_pathes = set()
        for obj in self.annotate(
            has_pathes=models.Exists(
                RulePath.objects.filter(
                    is_active=True, rule_id=models.OuterRef("rule_id")
                )
            )
        ):
            # we always annotate with a set, to differ from _empty
            set_ob = rule_id_to_names.setdefault(obj.rule_id, set())
            if obj.is_active:
                if obj.has_pathes:
                    name_has_pathes.add(obj.name)
                set_ob.add(obj.name)
        return rule_id_to_names, name_has_pathes

    aname_matchers = sync_to_async(name_matchers)


class RuleRatelimitGroup(ActivatableAndManageable):
    rule = models.ForeignKey(
        Rule, related_name="ratelimit_groups", on_delete=models.CASCADE
    )
    name = models.CharField(
        max_length=80,
        verbose_name="ratelimit group name",
        help_text="matcher for django-fast-ratelimit group",
        validators=[min_length_1],
    )

    objects = RuleRatelimitGroupManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["rule", "name"], name="rule_ratelimit_groups_unique"
            )
        ]


class RuleNetwork(ActivatableAndManageable):
    rule = models.ForeignKey(Rule, related_name="networks", on_delete=models.CASCADE)
    network = models.CharField(max_length=50, validators=[validate_network])

    def __str__(self):
        return self.network


class RuleSourceManager(models.Manager):
    def clear_remote_caches(
        self,
    ):
        cache = caches[getattr(settings, "IPRESTRICT_CACHE", "default")]
        cache.delete_many(
            self.annotate_with_cache_key().values_list("cache_key", flat=True),
        )

    def annotate_with_cache_key(self, queryset=None):
        if queryset is None:
            queryset = self.get_queryset()
        # last is rule id for easy extraction
        return queryset.annotate(
            cache_key=Concat(
                models.Value(
                    f'{getattr(settings, "IPRESTRICT_KEY_PREFIX", "fip:")}source_data:'
                ),
                models.F("id"),
                models.Value(":"),
                models.F("rule_id"),
                output_field=models.TextField(),
            ),
            force_expire_cache_key=Concat(
                models.Value(
                    f'{getattr(settings, "IPRESTRICT_KEY_PREFIX", "fip:")}source_force_expire:'
                ),
                models.F("interval"),
                models.Value(":"),
                models.F("rule_id"),
                output_field=models.TextField(),
            ),
        )

    def _is_force_expired(self, query, cache, set_all):
        if set_all:
            return set(query.values_list("cache_key", flat=True))
        expire_dates = cache.get_many(
            set(query.values_list("force_expire_cache_key", flat=True))
        )
        cur_time = int(time.time())
        expired = set()
        for source in query:
            if (
                source.force_expire_cache_key not in expire_dates
                or expire_dates[source.force_expire_cache_key] < cur_time
            ):
                expired.add(source.cache_key)
        return expired

    def ip_matchers_remote(self, rules):
        cache = caches[getattr(settings, "IPRESTRICT_CACHE", "default")]
        q = models.Q(is_active=True)
        if isinstance(rules, models.QuerySet):
            q &= models.Q(rule__in=rules)
        else:
            q &= models.Q(rule_id__in=rules)
        keys_query = self.annotate_with_cache_key(self.filter(q))
        keys = keys_query.values_list("cache_key", flat=True)
        max_interval = keys_query.aggregate(Max("interval", default=0))["interval__max"]
        force_expire = bool(getattr(settings, "IPRESTRICT_SOURCE_FORCE_EXPIRE", True))
        # <= force_expire_multiplier <= 0 disables force expire
        skip_cache = (
            set()
            if not force_expire
            else self._is_force_expired(keys_query, cache, max_interval == 0)
        )
        cache_result = (
            cache.get_many(set(keys).difference(skip_cache))
            if max_interval > 0 and not skip_cache
            else {}
        )
        last_interval = None
        to_set = {}
        result = {}

        for source in (
            keys_query.exclude(cache_key__in=cache_result.keys())
            .distinct()
            .order_by("interval")
        ):
            # FIXME: double calling, some kind of locking would be good
            networks = source.get_processed_networks_uncached()
            # annotated
            cache_key = source.cache_key
            force_expire_cache_key = source.force_expire_cache_key
            result.setdefault(cache_key.rsplit(":", 1)[-1], []).extend(networks)
            if last_interval is not None and last_interval != source.interval:
                # not empty
                if to_set:
                    cache.set_many(to_set, timeout=last_interval)
                to_set = {}
            # interval 0 disables caching
            if source.interval > 0:
                to_set[cache_key] = ",".join(map(lambda x: x.compressed, networks))
                if force_expire:
                    to_set[force_expire_cache_key] = int(time.time()) + source.interval

            last_interval = source.interval
        if last_interval is not None:
            # not empty
            if to_set:
                cache.set_many(to_set, timeout=last_interval)

        for cache_key, value in cache_result.items():
            if isinstance(value, str):
                result_key = cache_key.rsplit(":", 1)[-1]
                networks = result.setdefault(result_key, [])
                for network_str in value.split(","):
                    try:
                        networks.append(parse_ipnetwork(network_str))
                    except ValueError:
                        pass
        return result

    def __str__(self):
        return self.generator_fn


class RuleSource(ActivatableAndManageable):
    rule = models.ForeignKey(Rule, related_name="sources", on_delete=models.CASCADE)
    generator_fn = models.CharField(
        max_length=200,
        validators=[
            validate_generator_fn,
        ],
    )
    interval = models.PositiveIntegerField(blank=True, default=get_default_interval)
    objects = RuleSourceManager()

    def get_cache_key(self):
        # last is rule id for easy extraction
        return f'{getattr(settings, "IPRESTRICT_KEY_PREFIX", "fip:")}source_data:{self.id}:{self.rule_id}'

    def get_force_expire_cache_key(self):
        return f'{getattr(settings, "IPRESTRICT_KEY_PREFIX", "fip:")}source_force_expire:{self.interval}:{self.rule_id}'

    def get_processed_networks_uncached(self):
        ret = []
        success = False
        try:
            validate_generator_fn(self.generator_fn)
            success = True
        except ValidationError:
            pass
        if success:
            try:
                ip_nets = import_string(self.generator_fn)()
            except ImportError as exc:
                logger.warning(
                    'could not import source generator_fn: "%s", rule name %s, source id %s',
                    self.generator_fn,
                    self.rule.name,
                    self.id,
                    exc_info=exc,
                )
                return []
            except Exception as exc:
                logger.error(
                    'source generator_fn "%s" failed, rule name %s, source id %s',
                    self.generator_fn,
                    self.rule.name,
                    self.id,
                    exc_info=exc,
                )
                return []
            for network in ip_nets:
                try:
                    ret.append(parse_ipnetwork(network))
                except ValueError:
                    pass

        return ret


class RulePathManager(models.Manager):
    @lru_cache(maxsize=1)
    def path_matchers(self):
        patterns = {}
        for obj in self.exclude(rule__action=RULE_ACTION.disabled.value).filter(
            is_active=True
        ):
            patterns.setdefault(obj.rule_id, []).append(obj.get_processed_path())
        for key in patterns.keys():
            patterns[key] = re.compile("|".join(patterns[key]))
        return patterns

    apath_matchers = sync_to_async(path_matchers)

    def match_ip_and_path(
        self,
        ip: str,
        path: str,
        method: Optional[str] = None,
        ratelimit_group: Optional[str] = None,
        rule_id=None,
        remote=True,
        # overwrite default action
        default_action=None,
    ):
        # ordered
        # with generic = pathless, non-catchall and disabled candidates are included
        if rule_id:
            candidate = Rule.objects.match_ip(
                ip,
                rule_id=rule_id,
                method=method,
                ratelimit_group=ratelimit_group,
                remote=remote,
                generic=True,
            )
            if candidate[0] is not None:
                candidates = [candidate]
        else:
            candidates = Rule.objects.match_all_ip(
                ip,
                method=method,
                ratelimit_group=ratelimit_group,
                remote=remote,
                generic=True,
            )
        _path_matchers = self.path_matchers()
        ratelimits = []
        for _rule_id, action, is_catch_all, _ratelimits in candidates:
            if action == RULE_ACTION.disabled:
                continue
            matcher = _path_matchers.get(_rule_id, None)
            # matcher is None, catchall for pathes (can be also a real catch all)
            if matcher and matcher.match(path):
                ratelimits.extend(_ratelimits)
                if action == RULE_ACTION.only_ratelimit:
                    continue
                return _rule_id, action, is_catch_all, ratelimits
        return None, get_default_action(default_action), False, ratelimits

    amatch_ip_and_path = sync_to_async(match_ip_and_path)


class RulePath(ActivatableAndManageable):
    rule = models.ForeignKey(Rule, related_name="pathes", on_delete=models.CASCADE)
    path = models.TextField(max_length=4096)
    is_regex = models.BooleanField(default=False, blank=True)

    objects = RulePathManager()

    class Meta:
        verbose_name_plural = "Rule Pathes"

    def __str__(self):
        return self.path

    def get_processed_path(self):
        return self.path if self.is_regex else re.escape(self.path)

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude=exclude)
        try:
            validate_regex(self.path) if self.is_regex else validate_path(self.path)
        except ValidationError as exc:
            if exclude and "path" in exclude:
                raise exc
            else:
                raise ValidationError({"path": exc})



# Generated by Django 5.0.1 on 2024-01-30 16:11

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_fast_iprestrict", "0007_ruleratelimitgroup_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="rule",
            name="managed_fields",
            field=models.JSONField(blank=True, default=list, editable=False),
        ),
        migrations.AddField(
            model_name="rulenetwork",
            name="managed_fields",
            field=models.JSONField(blank=True, default=list, editable=False),
        ),
        migrations.AddField(
            model_name="rulepath",
            name="managed_fields",
            field=models.JSONField(blank=True, default=list, editable=False),
        ),
        migrations.AddField(
            model_name="ruleratelimit",
            name="managed_fields",
            field=models.JSONField(blank=True, default=list, editable=False),
        ),
        migrations.AddField(
            model_name="ruleratelimitgroup",
            name="managed_fields",
            field=models.JSONField(blank=True, default=list, editable=False),
        ),
        migrations.AddField(
            model_name="rulesource",
            name="managed_fields",
            field=models.JSONField(blank=True, default=list, editable=False),
        ),
    ]



# Generated by Django 5.0.2 on 2024-02-07 11:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_fast_iprestrict', '0010_alter_ruleratelimit_rate'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='ruleratelimitgroup',
            name='rule_ratelimit_groups_unique',
        ),
        migrations.AddConstraint(
            model_name='ruleratelimitgroup',
            constraint=models.UniqueConstraint(fields=('rule', 'name'), name='rule_ratelimit_groups_unique'),
        ),
    ]



# Generated by Django 5.0.1 on 2024-01-03 11:12

import django_fast_iprestrict.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_fast_iprestrict", "0005_add_ratelimit_fix_generator_fn"),
    ]

    operations = [
        migrations.AddField(
            model_name="rule",
            name="invert_methods",
            field=models.BooleanField(
                blank=True,
                default=True,
                help_text="exclude instead of include http methods. Set and leave methods empty to match all",
            ),
        ),
        migrations.AddField(
            model_name="rule",
            name="methods",
            field=models.CharField(
                blank=True,
                default="",
                help_text="comma seperated http methods",
                max_length=100,
                validators=[django_fast_iprestrict.validators.validate_methods],
            ),
        ),
        migrations.AddField(
            model_name="ruleratelimit",
            name="action",
            field=models.SmallIntegerField(
                choices=[
                    (2, "normal"),
                    (1, "annotate only"),
                    (3, "clear cache for key"),
                    (4, "reset to before request"),
                ],
                default=2,
            ),
        ),
    ]
